# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import jsonhome
from jsonhome.tests import base


class DocumentTests(base.TestCase):

    def setUp(self):
        super(DocumentTests, self).setUp()
        self.doc = jsonhome.Document()

    def assertDocument(self, data):
        self.assertEqual({'resources': data}, self.doc.to_dict())

    def test_create_allow_delete(self):
        r = self.doc.add_resource('relation', allow_delete=True)

        self.assertTrue(r.allow_delete)
        self.assertDocument({'relation': {'hints': {'allow': ['DELETE']}}})

    def test_create_docs(self):
        r = self.doc.add_resource('relation', docs='doc-location')

        self.assertEqual('doc-location', r.docs)
        self.assertDocument({'relation': {'hints': {'docs': 'doc-location'}}})

    def test_accept_patch(self):
        f = 'application/json-patch'
        r = self.doc.add_resource('relation', accept_patch=[f])

        # setting accept_patch also sets the allow_patch flag
        self.assertEqual([f], r.accept_patch)
        self.assertTrue(r.allow_patch)

        self.assertDocument({'relation': {'hints': {'accept-patch': [f],
                                                    'allow': ['PATCH']}}})

    def test_accept_post(self):
        f = 'application/json'
        r = self.doc.add_resource('relation', accept_post=[f])

        # setting accept_post also sets the allow_post flag
        self.assertEqual([f], r.accept_post)
        self.assertTrue(r.allow_post)

        self.assertDocument({'relation': {'hints': {'accept-post': [f],
                                                    'allow': ['POST']}}})

    def test_accept_ranges(self):
        r = self.doc.add_resource('relation', accept_ranges=['bytes'])

        self.assertEqual(['bytes'], r.accept_ranges)
        d = {'relation': {'hints': {'accept-ranges': ['bytes']}}}
        self.assertDocument(d)

    def test_accept_prefer(self):
        r = self.doc.add_resource('relation', accept_prefer=['preference'])

        self.assertEqual(['preference'], r.accept_prefer)
        d = {'relation': {'hints': {'accept-prefer': ['preference']}}}
        self.assertDocument(d)

    def test_href(self):
        r = self.doc.add_resource('relation', href='href-value')

        self.assertEqual('href-value', r.href)
        self.assertEqual('href-value', r.get_uri())
        self.assertEqual('href-value', self.doc.get_uri('relation'))
        self.assertDocument({'relation': {'href': 'href-value'}})

    def test_raises_type_error_on_unknown_create(self):
        e = self.assertRaises(TypeError,
                              self.doc.add_resource,
                              'relation',
                              variablea='foo',
                              variableb='bar')

        s = str(e)

        self.assertIn('variablea', s)
        self.assertIn('variableb', s)

    def test_cant_install_resource_twice(self):
        r = self.doc.add_resource('relation')

        self.assertRaises(jsonhome.ResourceAlreadyExists,
                          self.doc.add_resource,
                          'relation')

        self.assertRaises(jsonhome.ResourceAlreadyExists,
                          self.doc.__setitem__,
                          'relation',
                          r)

    def test_simple_document_equality(self):
        d1 = jsonhome.Document()
        d1.add_resource('relation', allow_delete=True)

        d2 = jsonhome.Document()
        d2.add_resource('relation', allow_delete=True)

        self.assertEqual(d1, d2)

    def test_equality_len_difference(self):
        d1 = jsonhome.Document()
        d1.add_resource('relation', allow_delete=True)
        d1.add_resource('another', allow_delete=True)

        d2 = jsonhome.Document()
        d2.add_resource('relation', allow_delete=True)

        self.assertNotEqual(d1, d2)

    def test_equality_relation_difference(self):
        d1 = jsonhome.Document()
        d1.add_resource('relation', allow_delete=True)
        d1.add_resource('foo', allow_delete=True)

        d2 = jsonhome.Document()
        d2.add_resource('relation', allow_delete=True)
        d2.add_resource('bar', allow_delete=True)

        self.assertNotEqual(d1, d2)

    def test_equality_resource_difference(self):
        d1 = jsonhome.Document()
        d1.add_resource('relation', allow_delete=True)
        d1.add_resource('another', allow_delete=True)

        d2 = jsonhome.Document()
        d2.add_resource('relation', allow_delete=True)
        d2.add_resource('another', allow_delete=False)

        self.assertNotEqual(d1, d2)

    def test_from_dict(self):
        x1 = {'resources': {'relation': {'hints': {'allow': ['DELETE']}}}}
        d = jsonhome.Document.from_dict(x1)

        self.assertTrue(d['relation'].allow_delete)

        x2 = d.to_dict()
        self.assertEqual(x1, x2)

    def test_from_json(self):
        x1 = '{"resources": {"relation": {"hints": {"allow": ["DELETE"]}}}}'
        d = jsonhome.Document.from_json(x1)

        self.assertTrue(d['relation'].allow_delete)

        x2 = d.to_json()
        self.assertEqual(x1, x2)

    def test_create_with_absolute_uri(self):
        r = self.doc.add_resource('relation', uri='href-value')
        self.assertEqual('href-value', r.href)
        self.assertIsNone(r.href_template)
        self.assertDocument({'relation': {'href': 'href-value'}})

    def test_create_with_template_uri(self):
        r = self.doc.add_resource('relation',
                                  uri='/path{/param}',
                                  uri_vars={'param': 'foo'})

        self.assertIsNone(r.href)
        self.assertEqual('/path{/param}', r.href_template)
        self.assertEqual({'param': 'foo'}, r.href_vars)

        self.assertEqual('/path/val',
                         self.doc.get_uri('relation', param='val'))

        self.assertDocument({'relation': {'href-template': '/path{/param}',
                                          'href-vars': {'param': 'foo'}}})

    def test_create_not_enough_variables(self):
        self.assertRaises(jsonhome.MissingValues,
                          self.doc.add_resource,
                          'relation',
                          uri='/path{/param}')

    def test_ignore_extra_uri_vars(self):
        self.doc.add_resource('relation',
                              uri='/path{/param}',
                              uri_vars={'param': 'foo',
                                        'extra': 'vals',
                                        'are': 'ignored'})

        self.assertDocument({'relation': {'href-template': '/path{/param}',
                                          'href-vars': {'param': 'foo'}}})

    def test_ignore_unused_uri_vars(self):
        self.doc.add_resource('relation',
                              href='href-value',
                              uri_vars={'param': 'foo',
                                        'extra': 'vals',
                                        'are': 'ignored'})

        self.assertDocument({'relation': {'href': 'href-value'}})

    def test_unknown_resource(self):
        self.assertRaises(jsonhome.UnknownResource,
                          self.doc.get_uri,
                          'unknown',
                          key='val')

    def test_cant_set_invalid_resource_type(self):
        def _f(k, v):
            """Function is just rather that using __setitem__ directly."""
            self.doc[k] = v

        self.assertRaises(TypeError, _f, 'relation', 'somestring')
        self.assertRaises(TypeError, _f, 'relation', 42)
        self.assertRaises(TypeError, _f, 'relation', ['list', 'of', 'stuff'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import json

import uritemplate


__all__ = ['Document',
           'Resource',

           'MEDIA_TYPE',

           'JsonHomeException',
           'MissingValues',
           'UnknownResource',
           'ResourceAlreadyExists'
           ]


MEDIA_TYPE = 'application/json-home'


class JsonHomeException(Exception):
    """Base Exception class that all JSONHome exceptions inherit from."""


class MissingValues(JsonHomeException):
    """There is not enough information in the JSONHome document to finish."""


class UnknownResource(JsonHomeException):
    """There is no resource available with the specified relation."""


class ResourceAlreadyExists(JsonHomeException):
    """A resource with the specified relation already exists."""


def _allow_prop(method):

    def _allow_getter(self):
        return self.is_allowed(method)

    def _allow_setter(self, value):
        in_list = self.is_allowed(method)

        if value and not in_list:
            self.allow.append(method)
        if in_list and not value:
            self.hints['allow'] = [x for x in self.allow
                                   if x.upper() != method]

    return property(_allow_getter,
                    _allow_setter,
                    doc='Allow the %s method on this resource' % method)


def _item_prop(name, default=None, setdefault=None, hint=False):
    """Create a property that fetches a value from the dictionary.

    We implement getter, setter and deleter here. Whilst we may not want users
    to do all of those things it is not our job to prevent users from doing
    something stupid. Our goal is to provide helper functions and if people go
    and delete hints or force set multiple href methods that's not our fault.

    Note that set
    :param default: A value that is returned if nothing present in the object.
        This value will be ignored if setdefault is set.
    :param callable setdefault: A value set and returned if nothing present in
        the object. This value will take priority to default. Note that this is
        a callable so the set value will be the result of executing this
        function. This allows us to create new objects for default values.
    :param bool hint: True if this attribute exists in the hints dictionary.

    :rtype: property
    """

    def o(self):
        return self.hints if hint else self

    def _getter(self):
        if setdefault:
            return o(self).setdefault(name, setdefault())
        else:
            return o(self).get(name, default)

    def _setter(self, value):
        o(self)[name] = value

    def _deleter(self):
        o(self).pop(name, None)

    return property(_getter, _setter, _deleter)


class Resource(dict):
    """One resource that exists within a JSON home document."""

    href_vars = _item_prop('href-vars', setdefault=dict)
    """A indication for variables in the template to construct a URI."""

    href_template = _item_prop('href-template')
    """A templated URI link to a resource."""

    href = _item_prop('href')
    """A Direct URI Link to a resource."""

    hints = _item_prop('hints', setdefault=dict)
    """Additional hint information defined by the resource.

    Resource hints allow clients to find relevant information about
    interacting with a resource beforehand, as a means of optimising
    communications, as well as advertising available behaviours (e.g., to
    aid in laying out a user interface for consuming the API).

    Hints are just that - they are not a "contract", and are to only be
    taken as advisory.  The runtime behaviour of the resource always
    overrides hinted information.

    For example, a resource might hint that the PUT method is allowed on
    all "widget" resources.  This means that generally, the user has the
    ability to PUT to a particular resource, but a specific resource
    might reject a PUT based upon access control or other considerations.
    More fine-grained information might be gathered by interacting with
    the resource (e.g., via a GET), or by another resource "containing"
    it (such as a "widgets" collection) or describing it (e.g., one
    linked to it with a "describedby" link relation).
    """

    allow = _item_prop('allow', setdefault=list, hint=True)
    """HTTP Allow Methods for this resource.

    Hints the HTTP methods that the current client will be able to use to
    interact with the resource; equivalent to the Allow HTTP response
    header.
    """

    accept_patch = _item_prop('accept-patch', setdefault=list, hint=True)
    """Hints the PATCH request formats accepted by the resource.

    This is equivalent to the Accept-Patch HTTP response header.
    """

    accept_post = _item_prop('accept-post', setdefault=list, hint=True)
    """Hints the POST request formats accepted by the resource."""

    accept_prefer = _item_prop('accept-prefer', setdefault=list, hint=True)
    """Hints the preferences supported by the resource.

    Note that, as per that specifications, a preference can be ignored by the
    server.
    """

    accept_ranges = _item_prop('accept-ranges', setdefault=list, hint=True)
    """Hints the range-specifiers available to the client.

    This is equivalent to the Accept-Ranges HTTP response header.
    """

    docs = _item_prop('docs', hint=True)
    """The location for human-readable documentation for the resource."""

    def is_allowed(self, method):
        """Test if a HTTP method can be used with this resource.

        :param str method: a HTTP method string to find.

        :returns: bool or None if no hints are defined by the resource.
        """
        try:
            allowed = self.hints['allow']
        except KeyError:
            return None
        else:
            return method.upper() in (a.upper() for a in allowed)

    allow_delete = _allow_prop('DELETE')
    allow_get = _allow_prop('GET')
    allow_head = _allow_prop('HEAD')
    allow_options = _allow_prop('OPTIONS')
    allow_patch = _allow_prop('PATCH')
    allow_post = _allow_prop('POST')
    allow_put = _allow_prop('PUT')

    def get_uri(self, **kwargs):
        """Get an absolute URI for this resource.

        Fetch the absolute URI. If there is an absolute URI set on this
        resource that will be returned.

        If there is a templated URI then the variables in the template will be
        evaluated against the values passed in through keyword arguments.
        """
        if self.href:
            return self.href

        if self.href_template:
            return uritemplate.expand(self.href_template, **kwargs)

        raise MissingValues("Couldn't determine href from values in Resource")

    def set_uri(self, uri, **kwargs):
        """Set the URI on this resource based on its format.

        Selective set the href or href-template and href-vars depending on if
        the URI contains a templated form or not. If there are variables set in
        the URI then a relation name be passed as a kwarg as well.

        Passing an absolute URI::

            res.set_uri('/path/to/resource')

        Passing a templated URI::

            res.set_uri('/path/to/resource{/param}',
                        param='http://description/rel/param')

        :param str uri: The URI you want to set. This may contain variables.

        :raises jsonhome.MissingValues: If a templated URI is passed and there
            is not a relation passed as a keyword argument that matches the
            variable.
        """
        variables = uritemplate.URITemplate(uri).variables

        if variables:
            try:
                href_vars = dict((n, kwargs.pop(n))
                                 for v in variables
                                 for n in v.variable_names)
            except KeyError as e:
                msg = "Missing parameter %s from template" % str(e)
                raise MissingValues(msg)

            del self.href
            self.href_template = uri
            self.href_vars = href_vars

        else:
            self.href = uri
            del self.href_template
            del self.href_vars

    @classmethod
    def create(cls, **kwargs):
        """Create a new resource with specified values.

        A factory function that allows you to create a new Resource object with
        a number of attributes in one method.

        :param str href: A direct URI link to a resource.
        :param str href_template: A template from which a URI is determined.
        :param dict href_vars: The raw href variables that should be set.

        :param str uri: A templatable URI, of the form expected by
            :py:meth:`~jsonhome.Resource.set_uri`.
        :param dict uri_vars: The URI variables that are interpolated in the
            form expected by :py:meth:`~jsonhome.Resource.set_uri`.

        :param str docs: location for human-readable documentation.

        :param bool allow_delete: allow the DELETE method on resource.
        :param bool allow_get: allow the GET method on resource.
        :param bool allow_head: allow the HEAD method on resource.
        :param bool allow_options: allow the OPTIONS method on resource.
        :param bool allow_patch: allow the PATCH method on resource.
        :param bool allow_post: allow the POST method on resource.
        :param bool allow_put: allow the PUT method on resource.

        :param accept_patch: Hints the PATCH [RFC5789] request formats accepted
            by the resource for this client; equivalent to the Accept-Patch
            HTTP response header.
        :type accept_patch: list(str)

        :param accept_post: Hints the POST request formats accepted by the
            resource for this client.
        :type accept_post: list(str)

        :param accept_prefer: Hints the preferences supported by the resource.
            Note that, as per that specifications, a preference can be ignored
            by the server.
        :type accept_prefer: list(str)

        :param accept_ranges: Hints the range-specifiers available to the
            client for this resource; equivalent to the Accept-Ranges HTTP
            response header
        :type accept_ranges: list(str)

        :rtype: :py:class:`~jsonhome.Resource`.
        """
        # NOTE(jamielennox): keep the above parameter list in sync with the
        # Document.add_resource function below for better documentation.

        # before we start handle some SHOULD aspects of the specification to
        # try and make the resources as consistent with the spec as possible.
        if kwargs.get('accept_patch'):
            kwargs.setdefault('allow_patch', True)
        if kwargs.get('accept_post'):
            kwargs.setdefault('allow_post', True)

        # ensure that the URI is only handled in one way.
        uri = kwargs.pop('uri', None)
        uri_vars = kwargs.pop('uri_vars', {})

        # NOTE(jamielennox): we purposefully only check uri here, not uri_vars
        # so that you can keep a repository of uri_vars that are passed every
        # time and ignored if no uri parameter is passed.
        if sum([bool(uri),
                bool(kwargs.get('href')),
                bool(kwargs.get('href_template') or
                     kwargs.get('href_vars'))]) > 1:
            m = 'You should choose only one way to set the URI on a resource.'
            raise ValueError(m)

        r = cls()

        if uri:
            r.set_uri(uri, **uri_vars)

        for method in ('href',
                       'href_template',
                       'href_vars',

                       'docs',

                       'allow_delete',
                       'allow_get',
                       'allow_head',
                       'allow_options',
                       'allow_patch',
                       'allow_post',
                       'allow_put',

                       'accept_patch',
                       'accept_post',
                       'accept_prefer',
                       'accept_ranges'):
            value = kwargs.pop(method, None)
            if value is not None:
                setattr(r, method, value)

        if kwargs:
            msg = 'create got an unexpected argument: %s' % ', '.join(kwargs)
            raise TypeError(msg)

        return r


class Document(dict):
    """A model of a JSON Home document that can be manipulated."""

    resource_class = Resource
    """The class of resource that should be created."""

    def __setitem__(self, relation, value):
        if not isinstance(value, self.resource_class):
            raise TypeError('Can only set valid resources on Document')

        if relation in self:
            raise ResourceAlreadyExists(relation)

        super(Document, self).__setitem__(relation, value)

    def get_uri(self, relation, **kwargs):
        """Get an absolute URI for this resource.

        Fetch the absolute URI. If there is an absolute URI set on this
        resource that will be returned.

        If there is a templated URI then the variables in the template will be
        evaluated against the values passed in through keyword arguments.

        :param str relation: The relation to the resource you wish to get the
            URI for.
        """
        try:
            res = self[relation]
        except KeyError:
            raise UnknownResource(relation)

        return res.get_uri(**kwargs)

    def add_resource(self, relation, **kwargs):
        """Create a new resource on this document.

        Create a new resource with prefilled attributes and add it to the
        current document.

        :param str relation: The string relationship to this resource that will
            be used to identity resources.

        :param str href: A direct URI link to a resource.
        :param str href_template: A template from which a URI is determined.
        :param dict href_vars: The raw href variables that should be set.

        :param str uri: A templatable URI, of the form expected by
            :py:meth:`~jsonhome.Resource.set_uri`.
        :param dict uri_vars: The URI variables that are interpolated in the
            form expected by :py:meth:`~jsonhome.Resource.set_uri`.

        :param str docs: location for human-readable documentation.

        :param bool allow_delete: allow the DELETE method on resource.
        :param bool allow_get: allow the GET method on resource.
        :param bool allow_head: allow the HEAD method on resource.
        :param bool allow_options: allow the OPTIONS method on resource.
        :param bool allow_patch: allow the PATCH method on resource.
        :param bool allow_post: allow the POST method on resource.
        :param bool allow_put: allow the PUT method on resource.

        :param accept_patch: Hints the PATCH [RFC5789] request formats accepted
            by the resource for this client; equivalent to the Accept-Patch
            HTTP response header.
        :type accept_patch: list(str)

        :param accept_post: Hints the POST request formats accepted by the
            resource for this client.
        :type accept_post: list(str)

        :param accept_prefer: Hints the preferences supported by the resource.
            Note that, as per that specifications, a preference can be ignored
            by the server.
        :type accept_prefer: list(str)

        :param accept_ranges: Hints the range-specifiers available to the
            client for this resource; equivalent to the Accept-Ranges HTTP
            response header
        :type accept_ranges: list(str)

        :raises jsonhome.ResourceAlreadyExists: If there is already a relation
            with the requested name available on the document.

        :returns: The new resource. This is the same resource that was added to
            the document so a user may perform additional manipulation.
        :rtype: :py:class:`~jsonhome.Resource`.
        """
        r = self.resource_class.create(**kwargs)
        self[relation] = r
        return r

    def to_dict(self):
        """Convert the document into a serializable format.

        Convert the current document into a valid json-home format so that it
        can be serialized into JSON.

        :rtype: dict
        """
        return {'resources': copy.deepcopy(self)}

    @classmethod
    def from_dict(cls, data):
        """Create a json-home document from de-serialized data.

        Convert a dict that may have been received from an external site into
        a json-home document that can be manipulated and queried.

        :param dict data: The data to be converted.

        :rtype: :py:class:`~jsonhome.Document`
        """
        return cls(dict((relation, cls.resource_class(d))
                        for relation, d in data['resources'].items()))

    def to_json(self, **kwargs):
        """Convert the Document into JSON format.

        Serialize the json-home document into valid JSON so that it can be sent
        to users.

        :rtype: str
        """
        return json.dumps(self.to_dict(), **kwargs)

    @classmethod
    def from_json(cls, data):
        """Create a JSON home document from a JSON string.

        Take a string that was received from a remote service and load the JSON
        home document that describes its resources.

        :rtype: :py:class:`~jsonhome.Document`
        """
        return cls.from_dict(json.loads(data))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import jsonhome
from jsonhome.tests import base


class ResourceTests(base.TestCase):

    def setUp(self):
        super(ResourceTests, self).setUp()
        self.res = jsonhome.Resource()

    def assertResource(self, data):
        self.assertEqual(data, self.res)

    def test_allow(self):
        self.res.allow.extend(['GET', 'PATCH'])
        self.res.allow.append('POST')

        self.assertEqual(['GET', 'PATCH', 'POST'], self.res.allow)
        self.assertResource({'hints': {'allow': ['GET', 'PATCH', 'POST']}})

    def test_allow_methods(self):
        self.res.allow_get = True
        self.assertTrue(self.res.allow_get)
        self.assertFalse(self.res.allow_post)
        self.assertEqual(['GET'], self.res.allow)

        self.res.allow_post = True
        self.assertTrue(self.res.allow_post)
        self.assertEqual(['GET', 'POST'], self.res.allow)

        self.res.allow_get = False
        self.assertFalse(self.res.allow_get)
        self.assertEqual(['POST'], self.res.allow)

    def test_docs(self):
        self.res.docs = 'doc-location'

        self.assertEqual('doc-location', self.res.docs)
        self.assertResource({'hints': {'docs': 'doc-location'}})

    def test_accept_patch(self):
        f = 'application/json-patch'
        self.res.accept_patch.append(f)

        self.assertEqual([f], self.res.accept_patch)
        self.assertResource({'hints': {'accept-patch': [f]}})

    def test_accept_post(self):
        f = 'application/json'
        self.res.accept_post.append(f)

        self.assertEqual([f], self.res.accept_post)
        self.assertResource({'hints': {'accept-post': [f]}})

    def test_accept_ranges(self):
        self.res.accept_ranges.append('bytes')

        self.assertEqual(['bytes'], self.res.accept_ranges)
        self.assertResource({'hints': {'accept-ranges': ['bytes']}})

    def test_accept_prefer(self):
        self.res.accept_prefer.append('preference')

        self.assertEqual(['preference'], self.res.accept_prefer)
        self.assertResource({'hints': {'accept-prefer': ['preference']}})

    def test_href(self):
        self.res.href = 'href-value'

        self.assertEqual('href-value', self.res.href)
        self.assertEqual('href-value', self.res.get_uri())
        self.assertResource({'href': 'href-value'})

    def test_raises_type_error_on_unknown_create(self):
        e = self.assertRaises(TypeError,
                              jsonhome.Resource.create,
                              variablea='foo',
                              variableb='bar')

        s = str(e)

        self.assertIn('variablea', s)
        self.assertIn('variableb', s)

    def test_setting_full_uri(self):
        uri = '/path/to/resource'
        self.res.set_uri(uri)

        self.assertEqual(uri, self.res.href)
        self.assertResource({'href': uri})

    def test_setting_template_uri(self):
        uri = '/path/to/resource{/vara}?param={varb}'
        vara = 'http://url/describes/vara'
        varb = 'http://url/describes/varb'

        self.res.set_uri(uri, vara=vara, varb=varb)

        href_vars = {'vara': vara, 'varb': varb}
        self.assertEqual(uri, self.res.href_template)
        self.assertEqual(href_vars, self.res.href_vars)
        self.assertResource({'href-template': uri, 'href-vars': href_vars})

        self.assertEqual('/path/to/resource/foo?param=',
                         self.res.get_uri(vara='foo'))
        self.assertEqual('/path/to/resource/foo?param=bar',
                         self.res.get_uri(vara='foo', varb='bar'))
        self.assertEqual('/path/to/resource?param=bar',
                         self.res.get_uri(varb='bar'))

    def test_not_enough_template_variables(self):
        self.assertRaises(jsonhome.MissingValues,
                          self.res.set_uri,
                          '/path/to/resource{/vara}?param={varb}',
                          vara='http://url/describes/vara')

    def test_no_href(self):
        self.assertRaises(jsonhome.MissingValues, self.res.get_uri)

    def test_create_multiple_href(self):
        self.assertRaises(ValueError,
                          jsonhome.Resource.create,
                          href='href-value',
                          href_template='href-template')

        self.assertRaises(ValueError,
                          jsonhome.Resource.create,
                          href='href-value',
                          uri='uri-value')

        self.assertRaises(ValueError,
                          jsonhome.Resource.create,
                          uri='uri-value',
                          href_template='href-template')
